from requests import get,post,delete as htdelete
from urllib.parse import urljoin
from contextlib import closing
from json import dumps,loads
from sys import stderr
from time import sleep
from kblab.result import create_result
import kblab
import logging

MAX_ID=2**38
MAX_RETRIES=2
MAX_RETRY_WAIT=60

logging.getLogger('requests').setLevel(logging.CRITICAL)
import requests.packages.urllib3
requests.packages.urllib3.disable_warnings()

class HttpArchive(kblab.Archive):
    def __init__(self, root=None, base=None, auth=None):
        url=root
        self.url = url + ('/' if url[-1] != '/' else '')
        self.base = base or url

        if auth:
            if isinstance(auth, dict) and auth['type'] == 'basic':
                self.auth = (auth['user'], auth['pass'])
            elif isinstance(auth, tuple) and len(auth) == 2:
                self.auth = auth
            else:
                raise Exception('unsupported auth configuration')
        else:
            self.auth = None

        n_retries=0
        retry_wait=1
        while True:
            try:
                r = get(urljoin(self.url, 'base'), auth=self.auth, verify=kblab.VERIFY_CA)
            except Exception as e:
                if n_retries == MAX_RETRIES:
                    raise Exception('Max number of retries (%d) reached. %s' % (MAX_RETRIES, str(e)))

                print('Exception while connecting. Retry %d of %s in %d seconds. %s' % (n_retries+1, str(MAX_RETRIES), retry_wait, str(e)), file=stderr, flush=True)
            else:
                if r.status_code in [ 401, 403 ]:
                    raise Exception('Unauthorized. HTTP status = %d' % r.status_code)

                if r.status_code == 200:
                    self.server_base = r.text
                    break;

                if n_retries == MAX_RETRIES:
                    raise Exception('Max number of retries (%d) reached.')
                
                print('Unexpected HTTP status code = %d Retry %d of %s in %d seconds' % (r.status_code, n_retries+1, str(MAX_RETRIES), retry_wait), file=stderr, flush=True)

            n_retries += 1
            sleep(retry_wait)

            if 2*retry_wait < MAX_RETRY_WAIT:
                retry_wait *= 2


    def new(self, **kwargs):
        r = post(urljoin(self.url,'new'), params=kwargs, auth=self.auth, verify=kblab.VERIFY_CA, allow_redirects=False)
        
        if r.status_code != 201:
            raise Exception('expected HTTP status 201, but got %d' % r.status_code)

        url = self.url + r.headers['Location'].split('/')[-2] + '/'

        return (self._get_key(url),
                kblab.Package(
                    url,
                    mode='a',
                    auth=self.auth,
                    server_base=url.replace(self.url, self.base)))


    def ingest(self, package, key=None):
        files = { path:package.get_raw(path) for path in package }
        r = post(self.url + 'ingest', files=files, auth=self.auth, verify=kblab.VERIFY_CA)

        if r.status_code != 201:
            raise Exception('expected HTTP status 201, but got %d with content %s' % (r.status_code, r.text))

        url = r.headers['Location']

        return self._get_key(url)


    def get(self, key, mode='r'):
        try:
            return kblab.Package(
                    self.url + key + '/',
                    mode=mode,
                    base=urljoin(self.base, key + '/'),
                    auth=self.auth,
                    server_base=urljoin(self.server_base, key + '/'))
        except kblab.HttpNotFoundException:
            return None
        except:
            raise


    def delete(self, key, force=False):
        if force:
            p = self.get(key, mode='a')
            r = htdelete(self.url + key + '/', auth=self.auth, verify=kblab.VERIFY_CA)

            if r.status_code == 404:
                return False

            if r.status_code not in  [ 200, 202, 204 ]:
                raise Exception(f'Expected status code 200, 202 or 204, got {r.status_code}')

            return True
        else:
            raise Exception('Use the force (parameter)')


    def search(self, query, start=0, max=None):
        g = self._search_iter(query, start, max)
        i,r,c = next(g).split()

        return create_result(i, r, c, g, self)


    def _search_iter(self, query, start=0, max=None):
        params = { 'q': dumps(query), 'start': start }
        if max: params.update({ 'max': max })

        with closing(get(urljoin(self.url, 'search'),
            params=params,
            verify=kblab.VERIFY_CA,
            auth=self.auth,
            stream=True)) as r:

            if r.status_code == 200:
                 for key in r.raw:
                     yield key[:-1].decode('utf-8')
            else:
                raise Exception('Expected status 200, got %d' % r.status_code)


    def count(self, query, cats={}):
        with closing(get(urljoin(self.url, 'count'),
                         verify=kblab.VERIFY_CA,
                         params={ 'q': dumps(query), 'c': dumps(cats) },
                         auth=self.auth)) as r:
                
            if r.status_code == 200:
                return loads(r.text)
            else:
                raise Exception('Expected status 200, got %d' % r.status_code)


    def get_location(self, key, path):
        return self.url + key + '/' + path


    def keys(self):
        with closing(get(self.url + 'packages', auth=self.auth, stream=True, verify=kblab.VERIFY_CA)) as r:
            if r.status_code == 200:
                for key in r.raw:
                    yield key[:-1].decode('utf-8')
            elif r.status_code != 404:
                raise Exception('server returned status %d' % r.status_code)


    def __iter__(self):
        yield from self.keys()


    def items(self):
        for key in self:
            yield (key, self.get(key))


    def _get_key(self, url):
        if url.startswith(self.url):
            return url[len(self.url):].split('/')[0]

        raise Exception('url (%s) does not start with %s' % (url, self.url))


    def __contains__(self, key):
        with closing(get(self.url + key + '/', auth=self.auth, verify=kblab.VERIFY_CA)) as r:
            return r.status_code == 200


    def __getitem__(self, key):
        return self.get(key)




import kblab

class Archive:
    def __new__(cls, root=None, **kwargs):
        if root and (isinstance(root, list) or isinstance(root, kblab.Archive)):
            return super(Archive, kblab.MultiArchive).__new__(kblab.MultiArchive)
        elif root and root.startswith('http'):
            return super(Archive, kblab.HttpArchive).__new__(kblab.HttpArchive)
        else:
            return super(Archive, kblab.FileArchive).__new__(kblab.FileArchive)




#!/usr/bin/env python3

from sys import argv,exit
from json import load,dumps
from copy import copy,deepcopy

levels = [ 'Issue', 'Monograph', 'Part', 'Page', 'Area', 'Text' ]


def aggregate(structure, content, environment):
    ret = copy(environment)
    ret_content = []

    stack = [ structure ]
    while len(stack) > 0:
        element = stack.pop(0)

        for x in element if isinstance(element, list) else [ element ]:
            if x.get('@id', '') in content and content[x['@id']].get('content', '') != '':
                ret_content += [ content[x['@id']].get('content', '') ]

            if 'has_part' in x:
                stack.insert(0, x['has_part'])

    ret['content'] = ret_content

    return ret


def flerge(structure, content, meta, level='Text', ignore = []):
    ret = []
    ignore += [ 'has_part', 'content' ]

    stack = [ (deepcopy(meta), element) for element in structure ]
    while len(stack) != 0:
        environment,element = stack.pop(0)
        nenviron = copy(environment)
        nenviron.update({ key:value for key,value in element.items() if key not in ignore })        

        if element.get('@type', None) == level:
            ret += [ aggregate(element, content, nenviron) ]
        elif 'has_part' in element:
            for x in element['has_part'][::-1]:
                stack.insert(0, (nenviron, x))

    return ret

if __name__ == '__main__':
    if len(argv) < 4 or len(argv) > 6:
        print(f'usage: {argv[0]} <structure-file> <content-file> <meta-file> [level] [ignore]')
        exit(1)

    level = argv[4] if len(argv) > 4 else 'Text'
    ignore = argv[5].split(',') if len(argv) > 5 else []

    with open(argv[1], mode='rb') as sfile, open(argv[2], mode='rb') as cfile, open(argv[3], mode='rb') as mfile:
        structure = load(sfile)
        content = { x['@id']:x for x in load(cfile) }
        meta = load(mfile)

        print(dumps(flerge(structure, content, meta, level, ignore), indent=2))




#!/usr/bin/env python

from uuid import uuid4
from hashlib import md5
from re import split
from contextlib import contextmanager
from os import makedirs
from os.path import dirname,exists,abspath
from hashlib import sha1
from io import BytesIO
from urllib.request import urlopen
from random import random
from datetime import datetime
from re import match
from json import load,dumps
from copy import copy,deepcopy

TEMP_PREFIX='/tmp/kblab-temp-'

def __init__():
    pass


def get_temp_dirname():
    return TEMP_PREFIX + '%s/' % uuid4()


def convert(n, radix=32, pad_to=0, alphabet='0123456789bcdfghjklmnpqrstvxzBCDFGHJKLMNPQRSTVXZ'):
    ret = ''

    if radix > len(alphabet) or radix == 0:
        raise Exception('radix == 0 or radix > len(alphabet)')

    while n:
        n, rest = divmod(n, radix)
        ret = alphabet[rest] + ret

    while len(ret) < pad_to:
        ret = alphabet[0] + ret

    return ret or alphabet[0]


def valid_path(path):
    if path[0] in [ '/' ] or '../' in path:
        raise Exception('invalid path (%s)' % path)

    return path


def valid_file(path):
    if path[0] == '/' or '..' in path:
        raise Exception('invalid path (%s)' % path)

    return path


def valid_key(key):
    if '/' in key or '.' in key:
        raise Exception('invalid key (%s)' % key)

    return key


def timestamp():
    return datetime.utcnow().isoformat() + 'Z'


def dict_search(a, b):
    if not (isinstance(a, dict) or isinstance(b, dict)):
        return wildcard_match(a, b)
    
    for key in a:
        if key == '*':
            return any([ dict_search(a[key], x) for x in b.values() ])
        if key not in b:
            if a[key] != None:
                return False
        elif not dict_search(a[key], b[key]):
            return False

    return True


def wildcard_match(a, b):
    return a == b or a == '*'


def dict_values(d, path):
    return _dict_values(d, path.split('.'))


def _dict_values(d, path):
    if path == []:
        if isinstance(d, list):
            return set([ x for x in d if not isinstance(x, (list, dict, tuple)) ])
        elif not isinstance(d, (list, dict, tuple)):
            return set([d])
    else:
        if isinstance(d, (list, tuple)):
            s=set()
            for x in d:
                s.update(_dict_values(x, path))
            return s
        elif isinstance(d, dict) and path[0] in d:
            return _dict_values(d[path[0]], path[1:])
    
    return set()


def chunked(f, chunk_size=10*1024, max=None):
    pos,b = 0,None
    while b != b'' and b != '':
        b = f.read(chunk_size if not max else min(chunk_size, max - pos))
        #print('chunk %d' % len(b))
        pos += len(b)

        if b != b'' and b != '':
            yield b


def decode_range(srange):
    s = match('bytes=(\\d+)-(\\d*)', srange).groups()

    return ( int(s[0]), int(s[1]) if s[1] != '' else None )


def aggregate(structure, content, environment):
    ret = copy(environment)
    ret_content = []

    stack = [ structure ]
    while len(stack) > 0:
        element = stack.pop(0)

        for x in element if isinstance(element, list) else [ element ]:
            if x.get('@id', '') in content and content[x['@id']].get('content', '') != '':
                ret_content += [ content[x['@id']].get('content', '') ]

            if 'has_part' in x:
                stack.insert(0, x['has_part'])

    ret['content'] = ret_content

    return ret


def flerge(structure, content, meta, level='Text', ignore = []):
    ret = []
    ignore += [ 'has_part', 'content' ]

    stack = [ (deepcopy(meta), element) for element in structure ]
    while len(stack) != 0:
        environment,element = stack.pop(0)
        nenviron = copy(environment)
        nenviron.update({ key:value for key,value in element.items() if key not in ignore })        

        if element.get('@type', None) == level:
            ret += [ aggregate(element, content, nenviron) ]
        elif 'has_part' in element:
            for x in element['has_part'][::-1]:
                stack.insert(0, (nenviron, x))

    return ret

#if __name__ == '__main__':
#    if len(argv) < 4 or len(argv) > 6:
#        print(f'usage: {argv[0]} <structure-file> <content-file> <meta-file> [level] [ignore]')
#        exit(1)
#
#    level = argv[4] if len(argv) > 4 else 'Text'
#    ignore = argv[5].split(',') if len(argv) > 5 else []
#
#    with open(argv[1], mode='rb') as sfile, open(argv[2], mode='rb') as cfile, open(argv[3], mode='rb') as mfile:
#        structure = load(sfile)
#        content = { x['@id']:x for x in load(cfile) }
#        meta = load(mfile)
#
#        print(dumps(flerge(structure, content, meta, level, ignore), indent=2))




#!/usr/bin/env python

from lxml import etree
#from xml.etree import ElementTree as etree
from argparse import ArgumentParser
from os.path import abspath,dirname,isfile
from sys import stdin,stdout,stderr,argv
from os import getpid,makedirs,remove
from inspect import getfile,currentframe
from json import loads

def fixalto(i):
    parser = etree.XMLParser()
    basedir = dirname(abspath(getfile(currentframe())))

    # find, correct and archive all OCR files
    fixalto = etree.XSLT(etree.parse('%s/fixalto.xsl' % basedir, parser))
    fixed = fixalto(etree.parse(i, parser))
    root = fixed.getroot()

    for cb in root.xpath("//*[ local-name() = 'ComposedBlock' ]"):
        # HPOS, etc already exists?
        if cb.xpath('@HPOS'):
            continue

        c = [ 100000000, 100000000, 0, 0 ]

        for tb in cb.xpath("*[ local-name() = 'TextBlock' ]"):
            t = [ int(tb.attrib['HPOS']), int(tb.attrib['VPOS']), int(tb.attrib['WIDTH']), int(tb.attrib['HEIGHT']) ]

            if t[0] < c[0]: c[0] = t[0]
            if t[1] < c[1]: c[1] = t[1]
            if t[0]+t[2] > c[2]: c[2] = t[0]+t[2]
            if t[1]+t[3] > c[3]: c[3] = t[1]+t[3]

            cb.attrib['HPOS'] = str(c[0])
            cb.attrib['VPOS'] = str(c[1])
            cb.attrib['WIDTH'] = str(c[2] - c[0])
            cb.attrib['HEIGHT'] = str(c[3] - c[1])

    return fixed


def get_alto_content(i, level='paragraph'):
    # fix alto
    fixed = fixalto(i)

    # get stylesheet
    parser = etree.XMLParser()
    basedir = dirname(abspath(getfile(currentframe())))

    # find, correct and archive all OCR files
    alto_to_json = etree.XSLT(etree.parse('%s/alto_to_json.xsl' % basedir, parser))
    js = alto_to_json(fixed)
    j = loads(str(js))

    if level == 'paragraph':
        j['composedblocks'] = [ {
            'i':x['i'],
            'c':x['c'],
            'textblocks': [ { 'content':" ".join([ z['w'] for z in w['words']]) } for w in x['textblocks'] ]
        } for x in j['composedblocks'] ]
    elif level == '':
        ...
    
    return j





from contextlib import closing
from json import loads,dumps
from requests import head,get,post,delete,put
from os.path import join,basename,abspath,isdir
from kblab.utils import valid_path,valid_key,chunked
from kblab.exceptions import RangeNotSupported
from hashlib import sha256
from copy import deepcopy
from urllib.parse import urljoin,unquote
from re import compile
from os import listdir
import kblab.package
from urllib.parse import urljoin
from tempfile import TemporaryFile
from io import BytesIO
from htfile import open as htopen

VERSION = 0.1

class HttpPackage(kblab.Package):
    def __init__(self, url, mode='r', base=None, auth=None, server_base=None, label=None):
        self.url = url
        self._mode = mode
        self.auth = auth
        self.base = base or url
        self.server_base = server_base or url

        if mode in [ 'r', 'a' ]:
            with closing(get(self.url, headers={ 'Accept': 'application/json' }, auth=self.auth, verify=kblab.VERIFY_CA)) as r:
                #print(self.url)

                if r.status_code == 404:
                    raise kblab.HttpNotFoundException(r.text)
                elif r.status_code != 200:
                    raise Exception('%d %s' % (r.status_code, r.text))

                self._desc = loads(r.text)

                if mode is 'a' and self._desc['status'] == 'finalized':
                    raise Exception('package is finalized, use patch(...)')

                self._desc['files'] = { x['path']:x for x in self._desc['files'] }
        elif mode == 'w':
            raise Exception('mode \'w\' not supported for HttpPackage(), use mode \'a\' or HttpArchive.new(...)')
        else:
            raise Exception('unsupported mode (\'%s\')' % mode)


    def get_location(self, path):
        return self.url + path


    def get_raw(self, path, range=None):
        headers = {}

        if path not in self:
            raise Exception('%s does not exist in package' % path)

        if range and not (range[0] == 0 and not range[1]):
            headers['Range'] = 'bytes=%d-%s' % (range[0], str(int(range[1])) if range[1] else '')

        # @TODO potential resource leak
        r = get(self.get_location(path), stream=True, auth=self.auth, headers=headers, verify=kblab.VERIFY_CA)
        r.raw.decode_stream = True

        if range and 'bytes' not in r.headers.get('Accept-Ranges', ''):
            raise RangeNotSupported()

        if r.status_code not in [ 200, 206 ]:
            raise Exception('Unexpected status %d' % r.status_code)

        return r.raw


    def get_iter(self, path, chunk_size=10*1024, range=None):
        if path in self:
            return self._get_iter(
                        self.get_raw(path, range=range),
                        chunk_size=chunk_size,
                        max=range[1]-range[0] if range and range[1] else None)
        else:
            raise Exception('%s does not exist in package' % path)


    def _get_iter(self, raw, chunk_size=10*1024, max=None):
        with raw as f:
            yield from chunked(f, chunk_size=chunk_size, max=max)


    def read(self, path):
        return get(self.url + path, auth=self.auth, verify=kblab.VERIFY_CA).text


    def list(self):
        return list(self._desc['files'].keys())


    def description(self):
        ret = deepcopy(self._desc)
        server_base = self.server_base

        if self.base != server_base:
            ret['@id'] = ret['@id'].replace(server_base, self.base)

            for path in ret['files']:
                f = ret['files'][path]
                f['@id'] = f['@id'].replace(server_base, self.base)

        # de-dict
        ret['files'] = [ x for x in ret['files'].values() ]

        return ret


    def close(self):
        self._mode = 'r'


    def status(self):
        return self._desc['status']


    @property
    def label(self):
        return self._desc['label']


    def _reload(self):
        self._desc = loads(get(self.url, auth=self.auth, verify=kblab.VERIFY_CA).text)
        self._desc['files'] = { x['path']:x for x in self._desc['files'] }


    def __iter__(self):
        return iter(self.list())


    def __contains__(self, key):
        return key in self._desc['files']


    def __getitem__(self, key):
        return self._desc['files'][key]


    def __setitem__(self, key, value):
        self._desc['files'][key] = value


    def __str__(self):
        return dumps(self.description(), indent=4)


#    def __len__(self):
#        return len(self._desc['files'])


def do_hash(fname):
    h = sha256()
    with open(fname, mode='rb') as f:
        b=None
        while b != b'':
            b = f.read(100*1024)
            h.update(b)

    return 'SHA256:' + h.digest().hex()




import kblab

VERSION = 0.1

class Package:
    def __new__(cls, first=None, **kwargs):
        if first and isinstance(first, kblab.Package):
            return super(Package, kblab.MultiPackage).__new__(kblab.MultiPackage)
        if first and first.startswith('http'):
            return super(Package, kblab.HttpPackage).__new__(kblab.HttpPackage)
        else:
            return super(Package, kblab.FilePackage).__new__(kblab.FilePackage)




from kblab.package import Package
from kblab.httppackage import HttpPackage
from kblab.archive import Archive
from kblab.httparchive import HttpArchive
import kblab.utils

VERIFY_CA=True

name = "kblab"




import kblab
from typing import NamedTuple
from collections import namedtuple
from collections.abc import Iterator
from multiprocessing import Pool

Hit = namedtuple('Hit', [ 'key', 'package' ])

class Result(NamedTuple):
    start: int
    n: int
    m: int
    keys: Iterator
    hits: Iterator

    def __iter__(self):
        return self.keys

    def map(self, f, processes=None, ordered=True, multi=False):
        if multi:
            with Pool(processes=processes) as pool:
               for x in pool.imap(f, self) if ordered else pool.imap_unordered(f, self):
                    yield x
        else:
            for x in map(f, self):
                yield x

def create_result(start, n, m, key_iter, archive):
    return Result(
            start,
            n,
            m,
            key_iter,
            package_generator(key_iter, archive) if archive else iter([]))


def package_generator(keys, archive):
    for key in keys:
        yield Hit(key, archive.get(key))




#!/usr/bin/env python3

from distutils.core import setup
import setuptools

setup(
    name='kblab-client',
    version='0.0.7a0',
    description='KB lab client',
    author='Martin Malmsten',
    author_email='martin.malmsten@kb.se',
    url="https://github.com/kungbib/kblab",
    install_requires = [
        'requests',
        'pyyaml',
        'lxml',
        'htfile'
    ],
    packages=[ 'kblab' ]
)

